@@ -16,6 +16,7 @@
 #include <locale.h>
 #include <fcntl.h>
 #include <stdio.h>
+#include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #ifdef _WIN32
@@ -26,10 +27,16 @@
 #include <termios.h>
 #endif
 
+#include "xkbcommon/xkbcommon.h"
+
+#include "darray.h"
+#include "keymap.h"
 #include "test.h"
 #include "utils.h"
 #include "utils-paths.h"
 #include "src/keysym.h"
+#include "src/xkbcomp/rules.h"
+#include "src/utils-numbers.h"
 
 #include "tools/tools-common.h"
 
@@ -49,12 +56,16 @@ test_init(void)
 void
 print_detailed_state(struct xkb_state *state)
 {
-    fprintf(stderr, "  Layout: base: %d, latched: %d, locked: %d, effective: %u\n",
+    fprintf(stderr,
+            "  Layout: base: %"PRId32", latched: %"PRId32", locked: %"PRId32", "
+            "effective: %"PRIu32"\n",
             xkb_state_serialize_layout(state, XKB_STATE_LAYOUT_DEPRESSED),
             xkb_state_serialize_layout(state, XKB_STATE_LAYOUT_LATCHED),
             xkb_state_serialize_layout(state, XKB_STATE_LAYOUT_LOCKED),
             xkb_state_serialize_layout(state, XKB_STATE_LAYOUT_EFFECTIVE));
-    fprintf(stderr, "  Modifiers: base: 0x%x, latched: 0x%x, locked: 0x%x, effective: 0x%x\n",
+    fprintf(stderr,
+            "  Modifiers: base: %#"PRIx32", latched: %#"PRIx32", "
+            "locked: %#"PRIx32", effective: %#"PRIx32"\n",
             xkb_state_serialize_mods(state, XKB_STATE_MODS_DEPRESSED),
             xkb_state_serialize_mods(state, XKB_STATE_MODS_LATCHED),
             xkb_state_serialize_mods(state, XKB_STATE_MODS_LOCKED),
@@ -97,7 +108,6 @@ test_key_seq_va(struct xkb_keymap *keymap, va_list ap)
 
     const xkb_keysym_t *syms;
     xkb_keysym_t sym;
-    unsigned int nsyms, i;
     char ksbuf[XKB_KEYSYM_NAME_MAX_SIZE];
     const char *opstr = NULL;
 
@@ -122,7 +132,7 @@ test_key_seq_va(struct xkb_keymap *keymap, va_list ap)
                 goto fail;
         }
 
-        nsyms = xkb_state_key_get_syms(state, kc, &syms);
+        const int nsyms = xkb_state_key_get_syms(state, kc, &syms);
         if (nsyms == 1) {
             sym = xkb_state_key_get_one_sym(state, kc);
             syms = &sym;
@@ -134,11 +144,14 @@ test_key_seq_va(struct xkb_keymap *keymap, va_list ap)
             xkb_state_update_key(state, kc, XKB_KEY_UP);
 
 #if HAVE_TOOLS
-        tools_print_keycode_state("", state, NULL, kc, XKB_CONSUMED_MODE_XKB, PRINT_ALL_FIELDS);
+        tools_print_keycode_state("", state, NULL, kc,
+                                  (op == DOWN ? XKB_KEY_DOWN : XKB_KEY_UP),
+                                  XKB_CONSUMED_MODE_XKB,
+                                  PRINT_ALL_FIELDS | PRINT_UNILINE);
 #endif
-        fprintf(stderr, "op %-6s got %u syms for keycode %3u: [", opstr, nsyms, kc);
+        fprintf(stderr, "op %-6s got %d syms for keycode %3"PRIu32": [", opstr, nsyms, kc);
 
-        for (i = 0; i < nsyms; i++) {
+        for (int i = 0; i < nsyms; i++) {
             keysym = va_arg(ap, int);
             xkb_keysym_get_name(syms[i], ksbuf, sizeof(ksbuf));
             fprintf(stderr, "%s%s", (i != 0) ? ", " : "", ksbuf);
@@ -259,7 +272,7 @@ test_get_path(const char *path_rel)
     if (!srcdir)
         srcdir = ".";
 
-    if (is_absolute(path_rel))
+    if (is_absolute_path(path_rel))
         return strdup(path_rel);
 
     path = asprintf_safe("%s/test/data%s%s", srcdir,
@@ -272,58 +285,69 @@ test_get_path(const char *path_rel)
 }
 
 char *
-test_read_file(const char *path_rel)
+read_file(const char *path, FILE *file)
 {
-    struct stat info;
-    char *ret, *tmp, *path;
-    int fd;
-
-    path = test_get_path(path_rel);
-    if (!path)
+    if (!file)
         return NULL;
 
-    fd = open(path, O_RDONLY);
-    free(path);
+    const int fd = fileno(file);
     if (fd < 0)
         return NULL;
 
+    struct stat info;
     if (fstat(fd, &info) != 0) {
-        close(fd);
+        fclose(file);
         return NULL;
     }
 
-    ret = malloc(info.st_size + 1);
+    char* ret = malloc(info.st_size + 1);
     if (!ret) {
-        close(fd);
+        fclose(file);
         return NULL;
     }
 
-    size_t remaining = info.st_size;
-    ssize_t count;
-    tmp = ret;
-    while ((count = read(fd, tmp, remaining)) > 0) {
-        remaining -= count;
-        tmp += count;
-    }
-    ret[info.st_size] = '\0';
-    close(fd);
-
-    if (remaining != 0) {
+    const size_t size = info.st_size;
+    const size_t count = fread(ret, sizeof(*ret), size, file);
+    if (count != size) {
+        if (!feof(file))
+            printf("Error reading file %s: unexpected end of file\n", path);
+        else if (ferror(file))
+            perror("Error reading file");
+        fclose(file);
         free(ret);
         return NULL;
     }
+    ret[count] = '\0';
 
     return ret;
 }
 
+char *
+test_read_file(const char *path_rel)
+{
+    char *path = test_get_path(path_rel);
+    if (!path)
+        return NULL;
+
+    FILE* file = fopen(path, "rb");
+    char *ret = NULL;
+
+    if (!file)
+        goto out;
+
+    ret = read_file(path, file);
+
+out:
+    if (file)
+        fclose(file);
+    free(path);
+    return ret;
+}
+
 struct xkb_context *
 test_get_context(enum test_context_flags test_flags)
 {
-    enum xkb_context_flags ctx_flags;
-    struct xkb_context *ctx;
-    char *path;
-
-    ctx_flags = XKB_CONTEXT_NO_DEFAULT_INCLUDES;
+    enum xkb_context_flags ctx_flags = XKB_CONTEXT_NO_DEFAULT_INCLUDES;
     if (test_flags & CONTEXT_ALLOW_ENVIRONMENT_NAMES) {
         unsetenv("XKB_DEFAULT_RULES");
         unsetenv("XKB_DEFAULT_MODEL");
@@ -335,11 +359,11 @@ test_get_context(enum test_context_flags test_flags)
         ctx_flags |= XKB_CONTEXT_NO_ENVIRONMENT_NAMES;
     }
 
-    ctx = xkb_context_new(ctx_flags);
+    struct xkb_context * const ctx = xkb_context_new(ctx_flags);
     if (!ctx)
         return NULL;
 
-    path = test_get_path("");
+    char * const path = test_get_path("");
     if (!path) {
         xkb_context_unref(ctx);
         return NULL;
@@ -352,7 +376,8 @@ test_get_context(enum test_context_flags test_flags)
 }
 
 struct xkb_keymap *
-test_compile_file(struct xkb_context *context, const char *path_rel)
+test_compile_file(struct xkb_context *context, enum xkb_keymap_format format,
+                  const char *path_rel)
 {
     struct xkb_keymap *keymap;
     FILE *file;
@@ -370,8 +395,8 @@ test_compile_file(struct xkb_context *context, const char *path_rel)
     }
     assert(file != NULL);
 
-    keymap = xkb_keymap_new_from_file(context, file,
-                                      XKB_KEYMAP_FORMAT_TEXT_V1, 0);
+    keymap = xkb_keymap_new_from_file(context, file, format,
+                                      XKB_KEYMAP_COMPILE_NO_FLAGS);
     fclose(file);
 
     if (!keymap) {
@@ -387,12 +412,13 @@ test_compile_file(struct xkb_context *context, const char *path_rel)
 }
 
 struct xkb_keymap *
-test_compile_string(struct xkb_context *context, const char *string)
+test_compile_string(struct xkb_context *context, enum xkb_keymap_format format,
+                    const char *string)
 {
     struct xkb_keymap *keymap;
 
-    keymap = xkb_keymap_new_from_string(context, string,
-                                        XKB_KEYMAP_FORMAT_TEXT_V1, 0);
+    keymap = xkb_keymap_new_from_string(context, string, format,
+                                        XKB_KEYMAP_COMPILE_NO_FLAGS);
     if (!keymap) {
         fprintf(stderr, "Failed to compile string\n");
         return NULL;
@@ -402,12 +428,13 @@ test_compile_string(struct xkb_context *context, const char *string)
 }
 
 struct xkb_keymap *
-test_compile_buffer(struct xkb_context *context, const char *buf, size_t len)
+test_compile_buffer(struct xkb_context *context, enum xkb_keymap_format format,
+                    const char *buf, size_t len)
 {
     struct xkb_keymap *keymap;
 
-    keymap = xkb_keymap_new_from_buffer(context, buf, len,
-                                        XKB_KEYMAP_FORMAT_TEXT_V1, 0);
+    keymap = xkb_keymap_new_from_buffer(context, buf, len, format,
+                                        XKB_KEYMAP_COMPILE_NO_FLAGS);
     if (!keymap) {
         fprintf(stderr, "Failed to compile keymap from memory buffer\n");
         return NULL;
@@ -417,8 +444,8 @@ test_compile_buffer(struct xkb_context *context, const char *buf, size_t len)
 }
 
 struct xkb_keymap *
-test_compile_rules(struct xkb_context *context, const char *rules,
-                   const char *model, const char *layout,
+test_compile_rules(struct xkb_context *context, enum xkb_keymap_format format,
+                   const char *rules, const char *model, const char *layout,
                    const char *variant, const char *options)
 {
     struct xkb_keymap *keymap;
@@ -431,9 +458,11 @@ test_compile_rules(struct xkb_context *context, const char *rules,
     };
 
     if (!rules && !model && !layout && !variant && !options)
-        keymap = xkb_keymap_new_from_names(context, NULL, 0);
+        keymap = xkb_keymap_new_from_names2(context, NULL, format,
+                                            XKB_KEYMAP_COMPILE_NO_FLAGS);
     else
-        keymap = xkb_keymap_new_from_names(context, &rmlvo, 0);
+        keymap = xkb_keymap_new_from_names2(context, &rmlvo, format,
+                                            XKB_KEYMAP_COMPILE_NO_FLAGS);
 
     if (!keymap) {
         fprintf(stderr,
@@ -445,25 +474,224 @@ test_compile_rules(struct xkb_context *context, const char *rules,
     return keymap;
 }
 
+static struct xkb_rmlvo_builder *
+xkb_rules_names_to_rmlvo_builder(struct xkb_context *context,
+                                 const struct xkb_rule_names *names)
+{
+    struct xkb_rmlvo_builder *rmlvo =
+        xkb_rmlvo_builder_new(context, names->rules, names->model,
+                              XKB_RMLVO_BUILDER_NO_FLAGS);
+    if (!rmlvo) {
+        fprintf(stderr, "ERROR: xkb_rmlvo_builder_new() failed\n");
+        return NULL;
+    }
+
+    char buf[1024] = { 0 };
+
+    /* First parse options, and gather the layout-specific ones and
+     * append the others to the builder */
+    darray(darray_string) loptions = darray_new();
+    if (!isempty(names->options)) {
+        const char *o = names->options;
+        while (*o != '\0') {
+            /* Get option name */
+            const char *option = o;
+            while (*o != '\0' && *o != ',' &&
+                   *o != OPTIONS_GROUP_SPECIFIER_PREFIX) {
+                    o++;
+            };
+            const size_t len = o - option;
+            if (len >= sizeof(buf))
+                goto error;
+            memcpy(buf, option, len);
+            buf[len] = '\0';
+
+            /* Check if layout-specific option */
+            xkb_layout_index_t layout = XKB_LAYOUT_INVALID;
+            if (*o == OPTIONS_GROUP_SPECIFIER_PREFIX) {
+                o++;
+                int count = parse_dec_to_uint32_t(o, SIZE_MAX, &layout);
+                if (count > 0 && layout > 0 && layout <= XKB_MAX_GROUPS) {
+                    o += count;
+                    layout--;
+                } else {
+                    layout = XKB_LAYOUT_INVALID;
+                }
+                const char* const layout_index_end = o;
+                while (*o != '\0' && *o != ',') { o++; }
+                if (count <= 0 || layout_index_end != o)
+                    layout = XKB_LAYOUT_INVALID;
+            }
+
+            if (layout != XKB_LAYOUT_INVALID) {
+                /* Save as layout-specific option, to be added with layout */
+                char *opt = strdup(buf);
+                if (!opt)
+                    goto error;
+                darray_resize0(loptions, layout + 1);
+                darray_append(darray_item(loptions, layout), opt);
+            } else {
+                /* Append layout-agnostic option */
+                if (!xkb_rmlvo_builder_append_option(rmlvo, buf))
+                    goto error;
+            }
+
+            if (*o == ',')
+                o++;
+        }
+    }
+
+    if (!isempty(names->layout)) {
+        /* Process layout & variant together.
+         * Ignore variants without corresponding layout. */
+        const char *l = names->layout;
+        const char *v = names->variant;
+        if (!names->variant)
+            v = "";
+        xkb_layout_index_t layout_count = 0;
+        while (*l != '\0') {
+            const char *layout = l;
+            const char *variant = v;
+            char *start = buf;
+            size_t buf_size = sizeof(buf);
+            while (*l != '\0' && *l != ',') { l++; };
+            while (*v != '\0' && *v != ',') { v++; };
+
+            size_t len = l - layout;
+            if (len >= buf_size)
+                goto error;
+            memcpy(start, layout, len);
+            start[len] = '\0';
+            start += len;
+            buf_size -= len + 1;
+
+            len = v - variant;
+            if (len >= buf_size)
+                goto error;
+            ++start;
+            memcpy(start, variant, len);
+            start[len] = '\0';
+
+            /* Get layout-specific options */
+            char **opts = NULL;
+            size_t opts_count = 0;
+            if (layout_count < darray_size(loptions)) {
+                opts = darray_items(darray_item(loptions, layout_count));
+                opts_count = darray_size(darray_item(loptions, layout_count));
+            }
+
+            if (!xkb_rmlvo_builder_append_layout(rmlvo, buf, start,
+                                                 (const char* *) opts,
+                                                 opts_count))
+                goto error;
+
+            if (*l == ',')
+                l++;
+            if (*v == ',')
+                v++;
+
+            layout_count += 1;
+        }
+    }
+
+exit:
+    {
+        darray_string *opts;
+        darray_foreach(opts, loptions) {
+            char **opt;
+            darray_foreach(opt, *opts)
+                free(*opt);
+            darray_free(*opts);
+        }
+        darray_free(loptions);
+    }
+
+    return rmlvo;
+
+error:
+    fprintf(stderr, "ERROR: %s\n", __func__);
+    xkb_rmlvo_builder_unref(rmlvo);
+    rmlvo = NULL;
+    goto exit;
+}
+
+struct xkb_keymap *
+test_compile_rmlvo(struct xkb_context *context, enum xkb_keymap_format format,
+                   const char *rules, const char *model, const char *layout,
+                   const char *variant, const char *options)
+{
+    struct xkb_keymap *keymap = NULL;
+    const struct xkb_rule_names names = {
+        .rules = rules,
+        .model = model,
+        .layout = layout,
+        .variant = variant,
+        .options = options
+    };
+
+    struct xkb_rmlvo_builder *rmlvo = xkb_rules_names_to_rmlvo_builder(context,
+                                                                       &names);
+
+    if (!rmlvo) {
+        fprintf(stderr,
+                "Failed to create RMLVO builder: "
+                "'%s', '%s', '%s', '%s', '%s'\n",
+                rules, model, layout, variant, options);
+        return NULL;
+    }
+
+    keymap = xkb_keymap_new_from_rmlvo(rmlvo, format,
+                                       XKB_KEYMAP_COMPILE_NO_FLAGS);
+
+    xkb_rmlvo_builder_unref(rmlvo);
+
+    if (!keymap) {
+        fprintf(stderr,
+                "Failed to compile RMLVO from builder: "
+                "'%s', '%s', '%s', '%s', '%s'\n",
+                rules, model, layout, variant, options);
+    }
+
+    return keymap;
+}
+
 bool
-test_compile_output(struct xkb_context *ctx,
+test_compile_output(struct xkb_context *ctx, enum xkb_keymap_format input_format,
+                    enum xkb_keymap_format output_format,
                     test_compile_buffer_t compile_buffer,
                     void *compile_buffer_private, const char *test_title,
                     const char *keymap_str, size_t keymap_len,
                     const char *rel_path, bool update_output_files)
+{
+    return test_compile_output2(ctx, input_format, output_format,
+                                TEST_KEYMAP_SERIALIZE_FLAGS,
+                                compile_buffer, compile_buffer_private,
+                                test_title, keymap_str, keymap_len, rel_path,
+                                update_output_files);
+}
+
+bool
+test_compile_output2(struct xkb_context *ctx,
+                     enum xkb_keymap_format input_format,
+                     enum xkb_keymap_format output_format,
+                     enum xkb_keymap_serialize_flags serialize_flags,
+                     test_compile_buffer_t compile_buffer,
+                     void *compile_buffer_private, const char *test_title,
+                     const char *keymap_str, size_t keymap_len,
+                     const char *rel_path, bool update_output_files)
 {
     int success = true;
     fprintf(stderr, "*** %s ***\n", test_title);
 
-    struct xkb_keymap *keymap =
-        compile_buffer(ctx, keymap_str, keymap_len, compile_buffer_private);
+    struct xkb_keymap *keymap = compile_buffer(
+        ctx, input_format, keymap_str, keymap_len, compile_buffer_private
+    );
 
     if (!rel_path) {
         /* No path given: expect compilation failure */
         if (keymap) {
-            char *got =
-                xkb_keymap_get_as_string(keymap,
-                                         XKB_KEYMAP_USE_ORIGINAL_FORMAT);
+            char *got = xkb_keymap_get_as_string2(keymap, output_format,
+                                                  serialize_flags);
             xkb_keymap_unref(keymap);
             assert(got);
             fprintf(stderr,
@@ -478,7 +706,8 @@ test_compile_output(struct xkb_context *ctx,
         return false;
     }
 
-    char *got = xkb_keymap_get_as_string(keymap, XKB_KEYMAP_USE_ORIGINAL_FORMAT);
+    char *got = xkb_keymap_get_as_string2(keymap, output_format,
+                                          serialize_flags);
     if (!got) {
         fprintf(stderr, "Unexpected keymap serialization failure\n");
         return false;
@@ -486,7 +715,7 @@ test_compile_output(struct xkb_context *ctx,
 
     xkb_keymap_unref(keymap);
 
-    char *path = test_get_path(rel_path);
+    char* const path = test_get_path(rel_path);
     assert(path);
 
     if (update_output_files) {
@@ -497,16 +726,20 @@ test_compile_output(struct xkb_context *ctx,
         fclose(file);
     } else {
         fprintf(stderr, "Reading golden test output: %s\n", path);
-        char *expected = test_read_file(rel_path);
+        char* const expected = test_read_file(rel_path);
         assert(expected);
-        const char *label[2] = {"Golden test", "Roundtrip"};
+        static const char *label[2] = {"Golden test", "Roundtrip"};
+        bool test_round_trip =
+            (output_format == input_format ||
+             output_format == XKB_KEYMAP_USE_ORIGINAL_FORMAT);
         for (unsigned int k = 0; k < ARRAY_SIZE(label) && success; k++) {
             if (streq(expected, got)) {
                 fprintf(stderr, "%s succeeded.\n", label[k]);
-                if (k > 0)
-                    continue;
+                if (!test_round_trip)
+                    break;
                 /* Test round trip */
-                keymap = compile_buffer(ctx, expected, strlen(expected),
+                keymap = compile_buffer(ctx, input_format,
+                                        expected, strlen(expected),
                                         compile_buffer_private);
                 if (!keymap) {
                     fprintf(stderr,
@@ -515,14 +748,15 @@ test_compile_output(struct xkb_context *ctx,
                     break;
                 }
                 free(got);
-                got = xkb_keymap_get_as_string(keymap,
-                                               XKB_KEYMAP_USE_ORIGINAL_FORMAT);
+                got = xkb_keymap_get_as_string2(keymap, output_format,
+                                                serialize_flags);
                 if (!got) {
                     fprintf(stderr,
                             "Unexpected keymap roundtrip serialization failure\n");
                     success = false;
                 }
                 xkb_keymap_unref(keymap);
+                test_round_trip = false;
             } else {
                 fprintf(stderr,
                         "%s failed: dumped map differs from expected.\n",
